from haizelabs_api import HaizeLabs as HaizeLabsAPI
from haizelabs_api.types.judge_call_params import (
    ContentTestContentInputOutputMessage,
)

from typing import List


class JudgesService:
    def __init__(self, api_key: str, user_id: str, api_client: HaizeLabsAPI):
        self.api_client = api_client
        self.user_id = user_id
        self.headers = {"x-api-key": api_key}

    def call(
        self,
        messages: List[ContentTestContentInputOutputMessage],
        judge_ids: List[str],
    ):
        content = {
            "input_messages": messages,
            "user_id": self.user_id,
        }
        result = self.api_client.judges.call(
            judge_ids=judge_ids, extra_headers=self.headers, content=content
        )

        return result



import requests
import os


def init_haizelabs():
    api_key = os.environ.get("HAIZE_API_KEY")

    if api_key is None:
        raise Exception(
            "HAIZE_API_KEY not found. Generate an api key on https://platform.haizelabs.com/app/settings and save as HAIZE_API_KEY in your enviornment"
        )

    return api_key


def get_user_id(api_key: str, base_url: str):
    response = requests.get(f"{base_url}/keys/get_user_id/{api_key}")
    if response.status_code != 200:
        raise Exception(
            "Error retreiving User. Make sure your api key is a valid api key from https://platform.haizelabs.com/app/settings"
        )
    user = response.json()
    return user



from typing import Callable, List
from haizelabs_api import HaizeLabs as HaizeLabsAPI

import time
import warnings


class TestingService:
    def __init__(self, api_key: str, user_id: str, api_client: HaizeLabsAPI):
        self.api_client = api_client
        self.user_id = user_id
        self.headers = {"x-api-key": api_key}

    def red_team(
        self,
        behaviors: List[str],
        inference_function: Callable,
        judge_ids: List[str],
        user_id: str = None,
        name: str = None,
    ):
        print(f"Starting red teaming run ({len(behaviors)} behaviors)...")

        # Start red teaming job
        start_response = self.api_client.testing.start(
            test_data={
                "name": name,
                "user_id": user_id,
                "test_type": "RED_TEAMING",
                "behaviors": [{"description": b} for b in behaviors],
                "detector_ids": judge_ids,
            }
        )

        test_id = start_response.test_id

        if start_response.status == "ERROR":
            raise Exception("Error starting experiment run.")

        content_count = 0
        processed_content_ids = []
        update_content_response = []

        # Poll until red teaming run is complete
        while True:
            # Check for and upload awaiting response content
            with (
                warnings.catch_warnings()
            ):  # Ignore pydantic warning about content type
                warnings.simplefilter("ignore")
                update_response = self.api_client.testing.update(
                    test_id=test_id, contents=update_content_response
                )

            if update_response.status in ["COMPLETE", "ERROR"]:
                break

            # Generate responses to awaiting responses content
            update_content_response = []
            for content in update_response.contents:
                if (
                    content.status == "AWAITING_RESPONSE"
                    and content.id not in processed_content_ids
                ):
                    processed_content_ids.append(content.id)
                    content_output = inference_function(content.input_messages)
                    content.output_messages = [
                        {"role": "assistant", "content": content_output}
                    ]
                    content.status = "ANALYZING_RESPONSE"
                    content_count += 1
                    update_content_response.append(content)

            # Sleep for a small amount of time
            time.sleep(0.2)

        print(f"{content_count} inputs tested.")
        print("Red teaming run completed.")

    def experiment(
        self,
        dataset_id: str,
        inference_function: Callable,
        judge_ids: List[str],
        user_id: str = None,
        name: str = None,
    ):
        print(f"Starting experiment run (dataset id: {dataset_id})...")

        # Start red teaming job
        start_response = self.api_client.testing.start(
            test_data={
                "name": name,
                "user_id": user_id,
                "test_type": "EXPERIMENT",
                "dataset_id": dataset_id,
                "detector_ids": judge_ids,
            }
        )

        test_id = start_response.test_id

        if start_response.status == "ERROR":
            raise Exception("Error starting experiment run.")

        content_count = 0
        processed_content_ids = []
        update_content_response = []

        # Poll until red teaming run is complete
        while True:
            # Check for and upload awaiting response content
            with (
                warnings.catch_warnings()
            ):  # Ignore pydantic warning about content type
                warnings.simplefilter("ignore")
                update_response = self.api_client.testing.update(
                    test_id=test_id, contents=update_content_response
                )

            if update_response.status in ["COMPLETE", "ERROR"]:
                break

            # Generate responses to awaiting responses content
            update_content_response = []
            for content in update_response.contents:
                if (
                    content.status == "AWAITING_RESPONSE"
                    and content.id not in processed_content_ids
                ):
                    processed_content_ids.append(content.id)
                    content_output = inference_function(content.input_messages)
                    content.output_messages = [
                        {"role": "assistant", "content": content_output}
                    ]
                    content.status = "ANALYZING_RESPONSE"
                    content_count += 1
                    update_content_response.append(content)

            # Sleep for a small amount of time
            time.sleep(0.2)

        print(f"{content_count} inputs tested.")
        print("Experiment run completed.")



from haizelabs_api import HaizeLabs as HaizeLabsAPI
from haizelabs.auth import get_user_id

import haizelabs.services as services
import os


class HaizeClient:
    def __init__(self, base_url: str = None, api_key: str = None):
        if not base_url and not (base_url := os.environ.get("HAIZE_LABS_BASE_URL")):
            base_url = f"https://api.haizelabs.com"

        if not api_key and not (api_key := os.environ.get("HAIZE_LABS_API_KEY")):
            raise Exception(
                "HAIZE_LABS_API_KEY not found. Generate an api key on https://platform.haizelabs.com/app/settings and save as HAIZE_API_KEY in your enviornment"
            )

        self.user_id = get_user_id(api_key, base_url)
        self.base_url = base_url
        self.api_key = api_key
        self.api_client = HaizeLabsAPI(base_url=base_url)
        self.judges = services.JudgesService(
            api_key=api_key, user_id=self.user_id, api_client=self.api_client
        )
        self.testing = services.TestingService(
            api_key=api_key, user_id=self.user_id, api_client=self.api_client
        )



from haizelabs_api.types.judge_call_params import (
    ContentTestContentInputOutputMessage,
)
from typing import Literal, Optional, List, Any, Dict
from pydantic import BaseModel, Field
from datetime import datetime
from enum import Enum

import contextvars
import uuid
import json

AddType = Literal["INPUT", "OUTPUT"]


class Message(BaseModel):
    role: Literal["assistant", "user", "tool"]
    content: str
    name: Optional[str]


class SpanType(str, Enum):
    JUDGE = "JUDGE"
    APP = "APP"
    MODEL = "MODEL"
    FUNCTION = "FUNCTION"
    SCORER = "SCORER"


class EvaluatorResult(BaseModel):
    name: Optional[str] = None
    label: Optional[str] = None
    score: float


class LogBase(BaseModel): ...


# TODO: confirm that this works with new content objects
class Content(BaseModel):
    id: str = Field(default_factory=lambda: uuid.uuid4().hex)
    time: datetime = Field(default_factory=lambda: datetime.now())
    user_id: Optional[str] = None

    input_messages: Optional[List[ContentTestContentInputOutputMessage]] = Field(
        default_factory=list
    )
    output_messages: Optional[List[ContentTestContentInputOutputMessage]] = Field(
        default_factory=list
    )

    input_detections: Optional[List[EvaluatorResult]] = Field(default_factory=list)
    output_detections: Optional[List[EvaluatorResult]] = Field(default_factory=list)

    content_group_ids: Optional[List[str]] = Field(default_factory=list)

    start: Optional[datetime] = Field(default_factory=lambda: datetime.now())
    end: Optional[datetime] = Field(default_factory=lambda: datetime.now())


class Span(LogBase):
    name: str = None
    id: str = Field(default_factory=lambda: uuid.uuid4().hex)
    trace_id: str = Field(default_factory=lambda: uuid.uuid4().hex)
    parent_id: str = None
    user_id: Optional[str] = None

    caller_id: Optional[str] = None

    content: Optional[Content] = None

    span_type: Optional[SpanType] = SpanType.APP

    start: datetime = Field(default_factory=lambda: datetime.now())
    end: datetime = Field(default_factory=lambda: datetime.now())

    metadata: Optional[Dict[str, Any]] = {}
    tags: Optional[Dict[str, Any]] = {}

    _context_token: Optional[contextvars.Token] = None

    def get_content(self):
        if not self.content:
            self.content = Content(
                id=self.id,
                time=self.start,
                start=self.start,
                end=self.end,
                user_id=self.user_id,
            )
        return self.content

    def set_input(self, input: Any, role: str = "user"):
        content = self.get_content()
        content.input_messages.append({"role": role, "content": json.dumps(input)})

    def set_output(self, output: Any, role: str = "user"):
        content = self.get_content()
        content.output_messages.append({"role": role, "content": json.dumps(output)})

    def add_evaluation(
        self,
        score: float,
        name: Optional[str] = None,
        label: Optional[str] = None,
        type: AddType = "INPUT",
    ):
        content = self.get_content()
        data = {"name": name, "label": label, "score": score}
        if type == "INPUT":
            content.input_evaluations.append(EvaluatorResult(**data))
        else:
            content.output_evaluations.append(EvaluatorResult(**data))

    def add_tag(self, key: str, value: Any):
        self.tags[key] = value

    def add_metadata(self, key: str, value: Any):
        self.metadata[key] = value

    def set_token(self, token: contextvars.Token):
        self._context_token = token

    def get_token(self):
        return self._context_token

    def reset_token(self):
        self._context_token = None

    def create_child(self, name: str):
        return Span(
            name=name,
            trace_id=self.trace_id,
            parent_id=self.id,
        )

    def begin(self):
        self.start = datetime.now()

    def finish(self):
        self.end = datetime.now()


class Trace(LogBase):
    id: str = Field(default_factory=lambda: uuid.uuid4().hex)
    name: Optional[str] = None
    start: datetime = Field(default_factory=lambda: datetime.now())
    end: datetime = Field(default_factory=lambda: datetime.now())
    user_id: Optional[str] = None

    root_content: Optional[Content] = None

    @staticmethod
    def create_from_span(span: Span):
        return Trace(
            id=span.trace_id,
            start=span.start,
            name=span.name,
            end=span.end,
            user_id=span.user_id,
            root_content=span.content,
        )



from haizelabs.logger.types import SpanType, LogBase, Span, Trace
from haizelabs_api import HaizeLabs as HaizeLabsAPI
from typing import Callable, List, Optional
from threading import Thread, Event, RLock

from contextlib import contextmanager
from queue import Empty, Queue
from functools import wraps

import contextvars
import inspect
import signal
import atexit
import json
import uuid
import sys
import os


class BackgroundLogger:
    def __init__(
        self, base_url: Optional[str] = None, batch_size: int = 16, api_key: str = ""
    ):
        self.batch_size = batch_size
        self._queue: Queue = Queue()
        self._submission_thread: Thread = Thread(target=self.submit_logs, daemon=True)
        self._started = False
        self._start_thread_lock = RLock()
        self._is_shutdown: Event = Event()

        self.api_key = api_key
        self.base_url = base_url
        self.api_client = HaizeLabsAPI(base_url=base_url)

        signal.signal(signal.SIGINT, self.handle_signal)
        signal.signal(signal.SIGTERM, self.handle_signal)
        atexit.register(self.shutdown)

    def _start(self):
        if not self._started:
            with self._start_thread_lock:
                if not self._started:
                    self._submission_thread.start()
                    self._started = True

    def handle_signal(self, signum, frame):
        self.shutdown()

    def log(self, log: LogBase):
        self._start()
        try:
            self._queue.put_nowait(log)
        except:
            # TODO: catch errors
            ...

    # TODO: normal logging here
    def _log_input(self, log: LogBase):
        headers = {"x-api-key": self.api_key}
        try:
            self.api_client.monitoring.log(**log.model_dump(), extra_headers=headers)
        except Exception as e:
            print(f"Error logging span: {str(e)}")
            # TODO: retry logic

    def submit_logs(self):
        while not self._is_shutdown.is_set():
            batch_logs: List[LogBase] = []
            while len(batch_logs) < self.batch_size:
                try:
                    log = self._queue.get(block=False)
                except Empty:
                    break
                else:
                    batch_logs.append(log)

            if batch_logs:
                for log in batch_logs:
                    self._log_input(log)

    def shutdown(self):
        if not self._started:
            try:
                sys.exit(0)
            except SystemExit:
                ...
        if not self._is_shutdown.is_set():
            self._is_shutdown.set()
            while not self._queue.empty():
                try:
                    log = self._queue.get_nowait()
                    self._log_input(log)
                except Empty:
                    break
        if self._submission_thread.is_alive():
            self._submission_thread.join()


def load_logger(batch_size=16):
    api_url = os.environ.get("HAIZE_LABS_BASE_URL", "https://api.haizelabs.com")
    api_key = os.environ.get("HAIZE_LABS_API_KEY")

    if api_key is None:
        raise Exception(
            "HAIZE_LABS_API_KEY not found. Generate an api key on https://platform.haizelabs.com/app/settings and save as HAIZE_API_KEY in your enviornment"
        )

    _logger = BackgroundLogger(base_url=api_url, batch_size=batch_size, api_key=api_key)
    return _logger


def _get_default_args(func: Callable):
    defaults = inspect.getfullargspec(func).defaults

    if not defaults:
        return {}

    args = inspect.getfullargspec(func).args
    return dict(zip(args[-len(defaults) :], defaults))


def _generate_id():
    return uuid.uuid4().hex


class _Tracer:
    def __init__(self):
        self._background_logger: BackgroundLogger = load_logger()
        self._current_span: contextvars.ContextVar[Span] = contextvars.ContextVar(
            "current_span", default=Span(name="root")
        )

        self._root_id_lock = RLock()
        self._root_token: contextvars.Token = None
        self.root_id: Optional[str] = None

    def log(self, log: LogBase):
        self._background_logger.log(log)

    def _prepare_outputs(self, ret):
        output = ret
        try:
            json.dumps(ret)
        except Exception as e:
            output = ""

        return output

    def _prepare_inputs(self, func_sig, func_args, func_kwargs, default_args):
        bound_args = func_sig.bind(*func_args, **func_kwargs).arguments
        serialized = {**default_args, **bound_args}
        try:
            json.dumps(serialized)
        except Exception as e:
            serialized = ""

        return serialized

    def get_current_span(self, name: str):
        span = self._current_span.get()
        if span.name == "root" and not self.root_id:
            with self._root_id_lock:
                span.id = _generate_id()
                span.trace_id = _generate_id()
                self.root_id = span.id
        return span.create_child(name)

    @contextmanager
    def trace(self, name: str, log_trace: bool = True):
        span = self.get_current_span(name)
        token = self._current_span.set(span)

        span.set_token(token)
        span.begin()
        yield span
        span.finish()

        if span.parent_id == self.root_id:
            span.parent_id = None
            trace = Trace.create_from_span(span)
            if log_trace:
                self.log(trace)
            self.root_id = None

        if log_trace:
            self.log(span)
            self.log(span.content)

        try:
            self._current_span.reset(span.get_token())
        except:
            ...
        span.reset_token()


def reset_tracer():
    global _tracer
    _tracer = _Tracer()


reset_tracer()


def get_tracer() -> _Tracer:
    return _tracer


def trace(
    name: Optional[str] = None,
    log_trace=True,
    span_type: SpanType = SpanType.APP,
):
    def decorator(func: Callable):
        sig = inspect.signature(func)
        default_args = _get_default_args(func)

        @wraps(func)
        def wrapper_sync(*args, **kwargs):
            with _tracer.trace(name, log_trace) as span:
                f_inputs = _tracer._prepare_inputs(sig, args, kwargs, default_args)
                ret = func(*args, **kwargs)
                f_outputs = _tracer._prepare_outputs(ret)
                span.span_type = span_type
                span.set_input(f_inputs)
                span.set_output(f_outputs)

                return ret

        @wraps(func)
        async def wrapper_async(*args, **kwargs):
            with _tracer.trace(name, log_trace) as span:
                f_inputs = _tracer._prepare_inputs(sig, args, kwargs, default_args)
                ret = await func(*args, **kwargs)
                f_outputs = _tracer._prepare_outputs(ret)
                span.span_type = span_type
                span.set_input(f_inputs)
                span.set_output(f_outputs)

                return ret

        if not log_trace:
            return func

        if inspect.iscoroutinefunction(func):
            return wrapper_async
        else:
            return wrapper_sync

    return decorator
